from .io import load_data, def_aliases, save_data
from .train import train_and_test, gbreweight, kfolding, binning_reweight
from .config import X_LABELS
from utils.plotting_utils import plot_distributions, plot_correlation_matrix, plot_roc_curve, plot_classifier_output, plot_weight_distributions
import joblib
import numpy as np

def run_reweighting_pipeline(args, plotdir, weightsdir):
    """
    Main function to run the reweighting pipeline.

    Args:
        args (argparse.Namespace): Command line arguments containing paths, variables, and other parameters.
    """
    # Load data
    data, sweights = load_data(path=args.path_data, tree=args.tree_data, columns=args.vars, weights_col=args.sweights_name)
    mc, _ = load_data(path=args.path_mc, tree=args.tree_mc, columns=args.vars)

    # Plot correlations
    plot_correlation_matrix(df=data, columns=args.vars, weights=sweights, output_file=f"{plotdir}/corr_data.png", title="Data")
    plot_correlation_matrix(df=mc, columns=args.vars, weights=None, output_file=f"{plotdir}/corr_mc.png", title="MC")

    # Train and test split
    mc_train, mc_test, data_train, data_test, w_train, w_test = train_and_test(mc=mc, data=data, weights=sweights, columns=args.vars, test_size=args.test_size)
    plot_distributions(mc=mc_train, data=data_train, mc_weights=np.ones(len(mc_train)), data_weights=w_train, columns=args.vars, x_labels=X_LABELS, output_file=f"{plotdir}/input_features_training.png")
    plot_distributions(mc=mc_test, data=data_test, mc_weights=np.ones(len(mc_test)), data_weights=w_test, columns=args.vars, x_labels=X_LABELS, output_file=f"{plotdir}/input_features_testing.png")

    # Train the model
    gb, gb_weights = gbreweight(mc_train=mc_train, data_train=data_train, mc_test=mc_test, data_test=data_test, w_train=w_train, w_test=w_test, columns=args.vars, trials=args.trials, weightsdir=weightsdir)
    gb_weights_kfolding = kfolding(gb=gb, mc=mc, data=data, sweights=sweights, columns=args.vars, n_folds=args.n_folds, weightsdir=weightsdir)
    binning_weights = binning_reweight(mc_train=mc_train, data_train=data_train, mc_test=mc_test, w_train=w_train, columns=args.vars, n_bins=args.n_bins, n_neighs=args.n_neighs, weightsdir=weightsdir)

    # Distributions
    plot_distributions(mc=mc_test, data=data_test, mc_weights=gb_weights, data_weights=w_test, columns=args.vars, x_labels=X_LABELS, output_file=f"{plotdir}/gb_weighted.png")
    plot_distributions(mc=mc, data=data, mc_weights=gb_weights_kfolding, data_weights=sweights, columns=args.vars, x_labels=X_LABELS, output_file=f"{plotdir}/kfolding_weighted.png")
    plot_distributions(mc=mc_test, data=data_test, mc_weights=binning_weights, data_weights=w_test, columns=args.vars, x_labels=X_LABELS, output_file=f"{plotdir}/binning_weighted.png")

    # ROC curve
    scores_gb, scores_kfolding, scores_binning = plot_roc_curve(mc_test=mc_test, data_test=data_test, sweights_test=w_test, 
                                                                mc=mc, data=data, sweights=sweights,
                                                                gb_weights=gb_weights, gb_weights_kfolding=gb_weights_kfolding, binning_weights=binning_weights, 
                                                                columns=args.vars, output_file=f"{plotdir}/roc_curve.png")

    # Draw classifier output
    plot_classifier_output(
        score_dict={"Data": scores_gb["Data"], "MC GBReweighted": scores_gb["MC"], "MC k-folding GBReweighted": scores_kfolding["MC"], "MC bins-reweighted": scores_binning["MC"]},
        weights_dict={"Data": w_test, "MC GBReweighted": gb_weights, "MC k-folding GBReweighted": gb_weights_kfolding, "MC bins-reweighted": binning_weights},
        min_score=0.1, max_score=0.4, output_file=f"{plotdir}/classifier_output.png"
    )
    
    # Plot weight distributions
    plot_weight_distributions(
        weights_dict={
            "GB": gb_weights,
            "KFold": gb_weights_kfolding,
            "Bins": binning_weights,
        },
        output_file=f"{plotdir}/weight_distributions.png",
        xlim=(0, 10)
    )

def apply_weights_pipeline(args, plotdir, weightsdir):
    """
    Main function to apply weights to the data using the trained model.

    Args:
        args (argparse.Namespace): Command line arguments containing paths, variables, and other parameters.
    """
    # Load data
    mc, _ = load_data(path=args.path_mc, tree=args.tree_mc, columns=args.vars)

    # Define aliases for training variables
    # Each alias should match the training variables used in the classifier with the name of the branch in the mc
    if(len(args.vars) != len(args.training_vars)):
        raise ValueError("The number of variables for reweighting must match the number of training variables."
                        " Please check the --vars and --training_vars arguments.")

    aliases = dict(zip(args.training_vars, args.vars))

    mc = def_aliases(mc, aliases)

    # Load method and weights
    if args.method == "gbreweight":
     classifier = joblib.load(f"{weightsdir}/gb_model_" + "_".join(args.vars) + ".pkl")
    elif args.method == "kfolding":
        classifier = joblib.load(f"{weightsdir}/gb_kfolding_model_" + "_".join(args.vars) + ".pkl") 
    elif args.method == "binning":
        classifier = joblib.load(f"{weightsdir}/binning_model_" + "_".join(args.vars) + ".pkl")
    else:
        raise ValueError(f"Unknown model type: {args.method}. Currently supported methods are 'gbreweight', 'kfolding', and 'binning'.")
    
    # Apply weights
    new_mc_weights = classifier.predict_weights(mc[args.vars])
    joblib.dump(new_mc_weights, f"{weightsdir}/mcweights_" + "_".join(args.vars) + ".pkl")

    # Write to output ROOT file
    weights_array = np.array(new_mc_weights, dtype=np.float32)
    save_data(input_path=args.path_mc, tree=args.tree_mc, output_path=args.output_path, output_tree=args.output_tree, branch=args.weights_name, weights=weights_array)
    