from hep_ml import reweight
from sklearn.model_selection import train_test_split
import joblib
from utils.utils import run_optuna
import numpy as np

def train_and_test(mc, data, weights, columns, test_size):
    """
    Split the data into training and testing sets.

    Args:
        mc (pd.DataFrame): MC data.
        data (pd.DataFrame): Data to reweight to.
        weights (np.ndarray): Sweights for the data.
        columns (list): List of column names to use for training.
        test_size (float): Proportion of the dataset to include in the test split.
    
    Returns:
        mc_train, mc_test, data_train, data_test, w_train, w_test: Split datasets and weights.
    """
    mc_train, mc_test = train_test_split(mc[columns], test_size=test_size, random_state=42)
    data_train, data_test, w_train, w_test = train_test_split(data[columns], weights, test_size=test_size, random_state=42)
    return mc_train, mc_test, data_train, data_test, w_train, w_test

def gbreweight(mc_train, data_train, mc_test, data_test, w_train, w_test, columns, trials, weightsdir):
    """
    Optimize hyperparameters using Optuna and train the Gradient Boosting Reweighter.

    Args:
        mc_train (pd.DataFrame): MC data for training.
        data_train (pd.DataFrame): Data to reweight to.
        w_train (np.ndarray): Sweights for the data.
        mc_test (pd.DataFrame): MC data for testing.
        data_test (pd.DataFrame): Data to reweight to for testing.
        w_test (np.ndarray): Sweights for the test data.
        columns (list): List of column names to use for training.
        trials (int): Number of trials for hyperparameter optimization.
        weightsdir (str): Directory to save the weights and model.
        test_size (float): Proportion of the dataset to include in the test split.
    Returns:
        gb (reweight.GBReweighter): Trained Gradient Boosting Reweighter.
        gb_weights (np.ndarray): Weights predicted by the trained model.
    """
    study_path = f"{weightsdir}/mult_and_kin_optuna_study_" + "_".join(columns) + ".pkl"
    if joblib.os.path.exists(study_path):
        study = joblib.load(study_path)
    else:
        study = run_optuna(mc_train, data_train, mc_test, data_test, w_train, w_test, columns, trials)
        joblib.dump(study, study_path)

    best = study.best_params
    gb = reweight.GBReweighter(
        n_estimators=best["n_estimators"],
        learning_rate=best["learning_rate"],
        max_depth=best["max_depth"],
        min_samples_leaf=best["min_samples_leaf"],
        gb_args={"subsample": best["subsample"]}
    )
    gb.fit(mc_train, data_train, original_weight=np.ones(len(mc_train)), target_weight=w_train)
    new_mc_weights = gb.predict_weights(mc_test)
    joblib.dump(gb, f"{weightsdir}/gb_model_" + "_".join(columns) + ".pkl")
    joblib.dump(new_mc_weights, f"{weightsdir}/gb_weights_" + "_".join(columns) + ".pkl")
    return gb, new_mc_weights

def kfolding(gb, mc, data, sweights, columns, n_folds, weightsdir):
    """
    Perform k-folding reweighting using the trained Gradient Boosting model.

    Args:
        gb (reweight.GBReweighter): Trained Gradient Boosting Reweighter.
        mc (pd.DataFrame): MC data.
        data (pd.DataFrame): Data to reweight to.
        sweights (np.ndarray): Sweights for the data.
        columns (list): List of column names to use for reweighting.
        n_folds (int): Number of folds for k-folding.
        weightsdir (str): Directory to save the weights and model.
    
    Returns:
        gb_weights_kfolding (np.ndarray): Weights for the MC data after k-folding reweighting.
    """
    gb_kfolding = reweight.FoldingReweighter(gb, n_folds=n_folds)
    gb_kfolding.fit(mc[columns], data[columns], original_weight=np.ones(len(mc)), target_weight=sweights)
    gb_weights_kfolding = gb_kfolding.predict_weights(mc[columns])
    joblib.dump(gb_kfolding, f"{weightsdir}/gb_kfolding_model_" + "_".join(columns) + ".pkl")
    joblib.dump(gb_weights_kfolding, f"{weightsdir}/gb_weights_kfolding_" + "_".join(columns) + ".pkl")
    return gb_weights_kfolding

def binning_reweight(mc_train, data_train, mc_test, w_train, columns, n_bins, n_neighs, weightsdir):
    """
    Perform binning reweighting on the provided MC and Data samples.

    Args:
        mc_train (pd.DataFrame): MC data for training.
        data_train (pd.DataFrame): Data used for training.
        mc_test (pd.DataFrame): MC data for testing.
        w_train (np.ndarray): Sweights for the training data.
        columns (list): List of column names to use for reweighting.
        n_bins (int): Number of bins to use for reweighting.
        n_neighs (int): Number of nearest neighbors to consider for reweighting.
        weightsdir (str): Directory to save the weights and model.
    Returns:
        new_mc_weights (np.ndarray): Weights for the MC data after reweighting.
    """
    bins = reweight.BinsReweighter(n_bins=n_bins, n_neighs=n_neighs)
    bins.fit(mc_train[columns], data_train[columns], original_weight=np.ones(len(mc_train)), target_weight=w_train)
    new_mc_weights = bins.predict_weights(mc_test[columns])
    joblib.dump(bins, f"{weightsdir}/binning_model_" + "_".join(columns) + ".pkl")
    joblib.dump(new_mc_weights, f"{weightsdir}/binning_weights_" + "_".join(columns) + ".pkl")
    return new_mc_weights