import numpy as np
import matplotlib.pyplot as plt
import math
from hep_ml.metrics_utils import ks_2samp_weighted
from utils.utils import evaluate_reweighting, weighted_corr_matrix
import seaborn as sns

hist_settings = {"bins": 50, "density": True, "alpha": 0.7}

def plot_correlation_matrix(df, columns, weights, title, output_file):
    """
    Plot a correlation matrix for the given DataFrame columns.

    Args:
        df (pd.DataFrame): DataFrame containing the data.
        columns (list): List of column names to include in the correlation matrix.
        weights (np.ndarray, optional): Weights for the correlation calculation. If None, unweighted correlation is used.
        title (str): Title of the plot.
        output_file (str): Path to save the output plot.
    """
    corr = weighted_corr_matrix(df, columns, weights) if weights is not None else df[columns].corr()
    plt.figure(figsize=(16, 12))
    sns.heatmap(corr, annot=True, fmt=".2f", cmap="coolwarm", square=True,
                cbar_kws={"shrink": .75}, xticklabels=columns, yticklabels=columns)
    plt.title(title)
    plt.tight_layout()
    plt.savefig(output_file)
    plt.close()

def plot_distributions(mc, data, mc_weights, data_weights, columns, x_labels, output_file, x_edges=None):
    """
    Plot distributions of multiple columns from MC and Data samples, comparing them with weighted histograms.

    Args:
        mc (pd.DataFrame): DataFrame containing MC data.
        data (pd.DataFrame): DataFrame containing Data.
        mc_weights (np.ndarray): Weights for the MC data.
        data_weights (np.ndarray): Weights for the Data.
        columns (list): List of column names to plot.
        x_labels (dict): Dictionary mapping column names to x-axis labels.
        output_file (str): Path to save the output plot.
        x_edges (dict, optional): Dictionary mapping column names to custom bin edges. If None, default binning is used.
    """ 
    n_cols = 3 if len(columns) != 4 else 2
    if len(columns) >= 10:
        n_cols = 5 
    n_rows = math.ceil(len(columns) / n_cols)
    plt.figure(figsize=[8*n_cols, 7*n_rows])
    for idx, column in enumerate(columns, 1):
        xlim = np.percentile(np.hstack([data[column]]), [0.01, 99.99]) if x_edges is None else (x_edges[column][0], x_edges[column][-1])
        plt.subplot(n_rows, n_cols, idx)
        plt.hist(mc[column], weights=mc_weights, range=xlim, label="MC", **hist_settings)
        plt.hist(data[column], weights=data_weights, range=xlim, label="Data", **hist_settings)
        plt.xlabel(x_labels[column])
        plt.title(f"{column}")
        plt.legend()
        print(
            f"KS over {column} = ",
            ks_2samp_weighted(
                mc[column],
                data[column],
                weights1=mc_weights,
                weights2=data_weights
            )
        )
    plt.savefig(output_file, bbox_inches="tight")
    plt.close()

def plot_roc_curve(mc_test, mc, data_test, data, sweights_test, sweights, gb_weights, gb_weights_kfolding, binning_weights, columns, output_file):
    """
    Plot ROC curve for the different reweighting methods.
    
    Args:
        mc_test (pd.DataFrame): MC testing sample.
        mc (pd.DataFrame): MC sample.
        data_test (pd.DataFrame): Data testing sample.
        data (pd.DataFrame): Data sample.
        sweights_test (np.ndarray): Sweights for data testing sample.
        sweights (np.ndarray): Sweights for data sample.
        gb_weights (np.ndarray): Weights predicted by the trained model.
        gb_weights_kfolding (np.ndarray): Weights from k-folding reweighting.
        binning_weights (np.ndarray): Weights from binning reweighting.
        columns (list): List of column names to use for plotting.
        output_file (str): Path to save the output plot.
    
    Returns:
        scores_gb, scores_kfolding, scores_binning (dict): Dictionaries containing classifier scores for each method.
    """
    fig, ax = plt.subplots(figsize=(16, 12))

    evaluate_reweighting(
        mc_test[columns].values, data_test[columns].values,
        np.ones(len(mc_test)), sweights_test, 
        "Unweighted", ax, {"MC": None, "Data": None}
    )

    scores_gb = {}
    evaluate_reweighting(
        mc_test[columns].values, data_test[columns].values,
        gb_weights, sweights_test,
        "GBReweighter", ax, scores_gb
    )

    scores_kfolding = {}
    evaluate_reweighting(
        mc[columns].values, data[columns].values,
        gb_weights_kfolding, sweights,
        "k-Folding GBReweighter", ax, scores_kfolding
    )

    scores_binning = {}
    evaluate_reweighting(
        mc_test[columns].values, data_test[columns].values,
        binning_weights, sweights_test,
        "BinsReweighter", ax, scores_binning
    )

    ax.plot([0, 1], [0, 1], linestyle='--', color='gray', label='Random')
    ax.set_ylabel("True Positive Rate")
    ax.set_title("ROC Curve: Classifier distinguishing reweighted MC from Data")
    ax.legend(loc="lower right")
    plt.savefig(output_file)
    print(f"ROC curve saved to: {output_file}")
    plt.close()
    
    return scores_gb, scores_kfolding, scores_binning

def plot_classifier_output(score_dict, weights_dict, min_score, max_score, output_file):
    """
    Plot the classifier output distributions for different labels and add KS test values in legend.

    Args:
        score_dict (dict): Dictionary where keys are labels and values are lists of classifier scores.
        weights_dict (dict): Dictionary where keys are labels and values are weights for the scores.
        min_score (float): Minimum score to consider for the plot.
        max_score (float): Maximum score to consider for the plot.
        output_file (str): Path to save the output plot.
    """
    plt.figure(figsize=(16, 12))
    
    # Reference = Data
    score_data = score_dict.get("Data")
    weights_data = weights_dict.get("Data")

    for label, score in score_dict.items():
        w = weights_dict[label]
        if label == "Data":
            # Plot Data normally
            legend_label = "Data"
        else:
            # Compute KS vs. Data
            ks_val = ks_2samp_weighted(score, score_data, weights1=w, weights2=weights_data)
            legend_label = f"{label} (KS = {ks_val:.3f})"
        
        plt.hist(score, bins=50, density=True, weights=w, alpha=0.6, label=legend_label, range=(min_score, max_score))

    plt.xlabel("Classifier output")
    plt.ylabel("Density")
    plt.title("Classifier score distribution with KS values")
    plt.legend()
    plt.savefig(output_file, bbox_inches="tight")
    print(f"Classifier output plot saved to: {output_file}")
    plt.close()

def plot_weight_distributions(weights_dict, output_file, bins=50, xlim=None):
    """
    Plot histograms of weight distributions.

    Args:
        weights_dict (dict): Dictionary where keys are labels and values are arrays of weights.
        output_file (str): Output file path for the plot.
        bins (int): Number of histogram bins.
        xlim (tuple or None): Limit for the x-axis, e.g., (0, 5). Default: auto.
    """
    plt.figure(figsize=(10, 7))
    for label, weights in weights_dict.items():
        plt.hist(weights, bins=bins, density=True, alpha=0.6, label=label, range=xlim, histtype="stepfilled")
    
    plt.xlabel("weights")
    plt.ylabel("Density")
    plt.legend()
    if xlim:
        plt.xlim(xlim)
    plt.yscale("log")  # Helps visualize long tails
    plt.grid(True, which="both", linestyle="--", alpha=0.5)
    plt.tight_layout()
    plt.savefig(output_file)
    print(f"Weight distributions plot saved to: {output_file}")
    plt.close()